import inspect
import types
from anticipate.adapt import adapt, adapt_all, register_adapter, AdaptError, AdaptErrors
import sys
from functools import partial, update_wrapper
from itertools import izip

__all__ = []

class AnticipateTypeError(Exception):
    def __init__(self, message, errors=None):
        super(AnticipateTypeError, self).__init__(message)
        self.errors = errors

class anticipate_wrapper(object):
    """
    Callable that is returned when you decorate something with `anticipate`.

    Handles checking or adapting the return type and input parameters.
    """
    def __init__(self, func, returns, params, strict=False):
        self.func = func
        self.returns = returns
        self.params = params
        self.adapt = adapt
        self.strict = strict

        if isinstance(self.returns, list):
            # Return value is a list of items matching the first element
            if len(self.returns) == 1:
                self.returns = self.returns[0]
                self.adapt = adapt_all
            else:
                raise TypeError('An anticipated list can only contain one type')

        self.arg_names = [n for n in inspect.getargspec(func)[0]]
        self.param_adapters = {}
        for key, p in self.params.items():
            if isinstance(p, list):
                # Value is a list of items matching the first element
                if len(p) == 1:
                    self.param_adapters[key] = partial(adapt_all, to_cls=p[0])
                else:
                    raise TypeError('An anticipated list can only contain one type')
            else:
                self.param_adapters[key] = partial(adapt, to_cls=p)

        # Make this look like the original function
        update_wrapper(self, self.func)

    def __get__(self, instance, owner):
        """
        If `anticipate` is decrating a method, `anticipate_wrapper` will be a descriptor.

        `__get__` will be called in this case which gives us an opportunity to bind to
        the instance.
        """

        return partial(self.__call__, instance)

    def __unadapted__(self, *args, **kwargs):
        """
        Call the wrapped function without adapting.
        """
        return self.func(*args, **kwargs)

    def _adapt_param(self, key, val):
        """
        Adapt the value if an adapter is defined.
        """
        if key in self.param_adapters:
            try:
                return self.param_adapters[key](val)
            except (AdaptError, AdaptErrors) as e:
                if hasattr(e, 'errors'):
                    errors = e.errors
                else:
                    errors = [e]

                raise AnticipateTypeError('Input value %r for %s does not match anticipated type %r' % (type(val), key, self.params[key]), errors=errors)
        else:
            return val

    def __call__(self, *args, **kwargs):
        """
        Call the wrapped function, adapting if needed.
        """

        if args and self.arg_names:
            args = list(args)

            # Replace args inline that have adapters
            for i, (key, val) in enumerate(izip(self.arg_names, args)):
                args[i] = self._adapt_param(key, val)

        if kwargs and self.params:
            # Adapt all adaptable arguments
            # TODO Strict checking
            for key, val in kwargs.items():
                kwargs[key] = self._adapt_param(key, val)

        result = self.func(*args, **kwargs)

        if self.returns:
            errors = None
            try:
                return self.adapt(result, self.returns)
            except AdaptErrors as e:
                errors = e.errors
            except AdaptError as e:
                errors = [e]

            raise AnticipateTypeError('Return value %r does not match anticipated type %r' % (type(result), self.returns), errors=errors)
        elif self.strict:
            if result is not None:
                raise AnticipateTypeError('Return value %r does not match anticipated value of None' % (type(result),), errors=None)
            return None
        else:
            return result



class anticipate(object):
    """
    A decorator that defines what a function/method expects.

    :param returns: Defines the type of object this function is expected to return.
                    Python basetypes as well as any object is supported. `None` means
                    that nothing can be returned. Use a list (ex: `[MyClass]`) to denote
                    that an iterable of `MyClass` will be returned.
    :param params: A list of key/type pairs. Each key corresponds to a parameter of the wrapped
                   function. The type is the type of object the function expects for that parameter.

    Example::

        @anticipate(dict, id=int)
        def get_obj(id):
            return {
                'id' : id
            }


    """
    def __init__(self, returns=None, **params):
        self.returns = returns
        self.params = params

    def __call__(self, func):
        return anticipate_wrapper(func, self.returns, self.params)

class strictly_anticipate(object):
    """
    Like `anticipate` but does not allow extra arguments or return values if `returns` is `None`
    """
    def __init__(self, returns=None, **params):
        self.returns = returns
        self.params = params

    def __call__(self, func):
        return anticipate_wrapper(func, self.returns, self.params, strict=True)

class adapter(object):
    """
    A decorator that registers an adapter

    :param from_cls: The class to convert from.
    :param to_cls: The class to convert to.

    Example::

        @adapter(int, str)
        def to_str(input, to_cls):
            return str(input)

    """
    def __init__(self, from_cls, to_cls):
        self.from_cls = from_cls
        self.to_cls = to_cls

    def __call__(self, func):
        register_adapter(self.from_cls, self.to_cls, func)
        return func


#!/usr/bin/env python
import pytest
import sys
from os import path

test_dir = path.dirname(path.abspath(__file__))

sys.path.insert(0, path.dirname(test_dir))

if __name__ == '__main__':
    raise SystemExit(pytest.main())


import pytest
from anticipate import adapt, adapter, anticipate


def test_mro():
    class Foo(object):
        def __init__(self, data):
            self.data = data

    class Bar(Foo):
        pass

    @adapter(Foo, str)
    def to_string(obj, to_cls):
        return 'Adapted to string'

    bar = Bar('My string')

    s = adapt.adapt(bar, str)

    assert s == 'Adapted to string'
    assert s == 'Adapted to string'

    class Zip(object):
        def __init__(self, data):
            self.data = data

    class Zam(Zip):
        pass

    @adapter(Foo, Zip)
    def from_foo_to_zip(obj, to_cls):
        return 'Adapted to zip'

    s = adapt.adapt(bar, Zam)

    assert s == 'Adapted to zip'

def test_adapt_params():
    @anticipate(foo=str, bar=int)
    def test(foo, bar, zing):
        return foo, bar, zing

    class FizzBuzz(object):
        def __str__(self):
            return 'fizzbuzz'

        def __int__(self):
            return 22

    @adapter((basestring, int, float, FizzBuzz), (basestring, int, float))
    def to_int(obj, to_cls):
        return to_cls(obj)

    assert test(1, 2.3, 'fizz') == ('1', 2, 'fizz')
    assert test(1, 2.3, zing='fizz') == ('1', 2, 'fizz')
    assert test(1, bar=2.3, zing='fizz') == ('1', 2, 'fizz')
    assert test(foo=1, bar=2.3, zing='fizz') == ('1', 2, 'fizz')

    assert test('hi', '1', str) == ('hi', 1, str)

    a = FizzBuzz()
    assert test(a, a, a) == ('fizzbuzz', 22, a)


def test_unadapted():
    """
    Ensure we can call a method without it being Adapted
    """
    @anticipate(foo=str, bar=int)
    def test(foo, bar):
        return foo, bar

    assert test(1, '0') == ('1', 0)
    assert test.__unadapted__(1, '0') == (1, '0')


def test_bound_to():
    """
    Ensure that an anticipated method gets bound to the subclass,
    not it's base class.
    """
    class BaseClass(object):
        @anticipate()
        def get_wrapped_self(self):
            return self

    class SubClass(BaseClass):
        def get_self(self):
            return self
            


    a = BaseClass()
    assert a.get_wrapped_self() == a
    

    b = SubClass()
    assert b.get_wrapped_self() == b
    assert b.get_wrapped_self() == b.get_self()


def test_instance_bound_to():
    """
    Ensure that an anticipated method gets bound to each instance
    instead of its first instance.
    """

    class BaseClass(object):
        pass

    class SubClass(BaseClass):
        def __init__(self, *args, **kwargs):
            super(SubClass, self).__init__(*args, **kwargs)
            self.thing = {}

        @anticipate()
        def get_wrapped_self(self):
            return self

        def get_self(self):
            return self

    b = SubClass()
    assert b.get_wrapped_self() is b
    assert b.get_wrapped_self() is b.get_self()        

    assert id(b.get_wrapped_self().thing) == id(b.get_self().thing)


    c = SubClass()
    assert c.get_wrapped_self() is c
    assert c.get_wrapped_self() is c.get_self()  

    assert id(c.get_wrapped_self().thing) == id(c.get_self().thing)


def test_args():
    """
    Verify that `self` is the correct instance when
    additional positional arguments are passed in.
    """
    class Test(object):
        @anticipate()
        def get_args(self, arg, *args):
            return self, arg, args

        @anticipate()
        def get_arg(self, arg):
            return self, arg

        @anticipate()
        def get_self(self, foo=None):
            return self

        @anticipate(arg=int)
        def get_arg_int(self, arg):
            return self, arg

        @anticipate(arg=int)
        def get_args_int(self, arg, *args):
            return self, arg, args

    @anticipate(arg=int)
    def get_arg_int(arg):
        return arg

    obj1 = object()
    obj2 = object()
    obj3 = object()

    b = Test()

    assert b.get_self() is b

    # Verify that if there are no adapters, *args pass through
    r = b.get_args(obj1, obj2, obj3)
    
    assert r[0] is b
    assert r[1] is obj1
    assert r[2][0] is obj2
    assert r[2][1] is obj3

    # Verify that if there are adapters, positional args get adapted
    r = b.get_args_int('1', obj2, obj3)
    
    assert r[0] is b
    assert r[1] == 1
    assert r[2][0] is obj2
    assert r[2][1] is obj3

    # Verify that if there are no adapters, positional args pass through
    r = b.get_arg(obj1)

    assert r[0] is b
    assert r[1] is obj1

    # Verify that if there are no adapters, keyword args pass through
    r = b.get_arg(arg=obj1)

    assert r[0] is b
    assert r[1] is obj1

    # Verify that keyword args are adapted
    r = b.get_arg_int(arg='1')

    assert r[0] is b
    assert r[1] == 1

    assert get_arg_int(arg='1') == 1

def test_kwargs():
    """
    Verify that kwargs can be adapted
    """
    class Test(object):
        @anticipate(foo=int, bar=str)
        def get_args(self, arg, **kwargs):
            return arg, kwargs

    @anticipate(foo=int, bar=str)
    def get_args(arg, **kwargs):
        return arg, kwargs

    t = Test()

    obj = object()

    r = t.get_args(obj, foo='2', bar=3)
    assert r[0] is obj
    assert r[1]['foo'] == 2
    assert r[1]['bar'] == '3'

    r = get_args(obj, foo='2', bar=3)
    assert r[0] is obj
    assert r[1]['foo'] == 2
    assert r[1]['bar'] == '3'    

    r = t.get_args(arg=obj, foo='2', bar=3)
    assert r[0] is obj
    assert r[1]['foo'] == 2
    assert r[1]['bar'] == '3'

    r = get_args(arg=obj, foo='2', bar=3)
    assert r[0] is obj
    assert r[1]['foo'] == 2
    assert r[1]['bar'] == '3'    


def test_adapt_all_generator():
    """
    Verify adapt_all returns a generator
    """
    int_like = ['1', 2.0]
    
    generator = adapt.adapt_all(int_like, int)
    assert generator.next() == 1
    assert generator.next() == 2

    assert list(adapt.adapt_all(int_like, int)) == [1, 2]


def test_adapt_all_with_none():
    """
    Verify passing None to adapt_all returns an empty generator
    """
    generator = adapt.adapt_all(None, int)

    with pytest.raises(StopIteration):
        generator.next()

    assert list(adapt.adapt_all(None, int)) == []
    

def test_anticipate_list():
    """
    Verify using a list for a parameter adapts the input to a generator
    that adapts each value in the input.
    """
    @anticipate(items=[int])
    def get_list(items):
        return items

    int_like = ['1', 2.0]

    r = get_list(int_like)
    assert r.next() == 1
    assert r.next() == 2.0

    # Works on list input
    assert list(get_list(int_like)) == [1, 2]

    # Works on tuple input
    assert list(get_list((4.0, 5.0, 6.0))) == [4, 5, 6]

    # Works on generator input
    assert list(get_list(iter((4.0, 5.0, 6.0)))) == [4, 5, 6]


def test_anticipate_list_with_none():
    """
    Verify passing None for a parameter that expects a list returns an
    empty generator.
    """
    @anticipate(items=[int])
    def get_list(items):
        return items

    generator = get_list(None)
    with pytest.raises(StopIteration):
        generator.next()
    
    assert list(get_list(None)) == []




from distutils.core import setup

def main():
    setup(
        name = 'anticipate',
        packages=['anticipate'],
        package_dir = {'':'src'},
        version = open('VERSION.txt').read().strip(),
        author='Mike Thornton',
        author_email='six8@devdetails.com',
        url='http://github.com/six8/anticipate',
        download_url='http://github.com/six8/anticipate',
        keywords=['packaging'],
        license='BSD',
        description='A type checking and adapting library',
        classifiers = [
            "Programming Language :: Python",
            "Intended Audience :: Developers",
            "Natural Language :: English",
            "Topic :: Software Development :: Libraries :: Python Modules",
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.5',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Operating System :: OS Independent',
            'License :: OSI Approved :: BSD License',
            "Development Status :: 3 - Alpha",
        ],
        long_description=open('README.rst').read(),
    )

if __name__ == '__main__':
    main()


import inspect
import itertools
import sys
import traceback

__adapters__ = {}
__mro__ = {}
__all__ = [
    'AdaptError',
    'AdapterExists',
    'adapt',
    'adapt_all',
    'register_adapter',
]

class AdaptError(Exception):
    pass

class AdaptErrors(AdaptError):
    def __init__(self, message, errors=None):
        super(AdaptErrors, self).__init__(message)
        self.errors = []
        if errors:
            self.add_errors(errors)

    def add_error(self, func, ex_type, ex, tb):
        self.errors.append((func, ex_type, ex, tb))

    def add_errors(self, errors):
        for e in errors:
            self.add_error(*e)

    def errors_string(self):
        """
        Returns all errors as a string
        """
        output = []
        for e in self.errors:
            output.append('%s: %s in %s:' % (e[1], e[2], e[0]))
            output.append(''.join(traceback.format_tb(e[3])))
        return '\n'.join(output)

class AdapterExists(Exception):
    pass

def get_adapter_path(obj, to_cls):
    """
    Returns the adapter path that would be used to adapt `obj` to `to_cls`.
    """
    from_cls = type(obj)
    key = (from_cls, to_cls)
    if key not in __mro__:
        __mro__[key] = list(itertools.product(inspect.getmro(from_cls), inspect.getmro(to_cls)))

    return __mro__[key]

def adapt(obj, to_cls):
    """
    Will adapt `obj` to an instance of `to_cls`.

    First sees if `obj` has an `__adapt__` method and uses it to adapt. If that fails
    it checks if `to_cls` has an `__adapt__` classmethod and uses it to adapt. IF that
    fails, MRO is used. If that
    fails, a `TypeError` is raised.
    """
    if obj is None:
        return obj
    elif isinstance(obj, to_cls):
        return obj

    errors = []

    if hasattr(obj, '__adapt__') and obj.__adapt__:
        try:
            return obj.__adapt__(to_cls)
        except (AdaptError, TypeError) as e:
            ex_type, ex, tb = sys.exc_info()
            errors.append((obj.__adapt__, ex_type, ex, tb))

    if hasattr(to_cls, '__adapt__') and to_cls.__adapt__:
        try:
            return to_cls.__adapt__(obj)
        except (AdaptError, TypeError) as e:
            ex_type, ex, tb = sys.exc_info()
            errors.append((to_cls.__adapt__, ex_type, ex, tb))

    for k in get_adapter_path(obj, to_cls):
        if k in __adapters__:
            try:
                return __adapters__[k](obj, to_cls)
            except (AdaptError, TypeError) as e:
                ex_type, ex, tb = sys.exc_info()
                errors.append((__adapters__[k], ex_type, ex, tb))
                break

    raise AdaptErrors('Could not adapt %r to %r' % (obj, to_cls), errors=errors)


def adapt_all(iterable, to_cls):
    """
    Returns a generator that will adapt all objects in an iterable to `cls`

    If `iterable` is `None`, an empty generator will be returned.
    """
    if iterable is None:
        # Use an empty tuple so we return an empty generator
        iterable = ()

    return (adapt(obj, to_cls) for obj in iterable)


def register_adapter(from_classes, to_classes, func):
    """
    Register a function that can handle adapting from `from_classes` to `to_classes`.
    """
    assert from_classes, 'Must supply classes to adapt from'
    assert to_classes, 'Must supply classes to adapt to'
    assert func, 'Must supply adapter function'

    if not isinstance(from_classes, (tuple, list)):
        from_classes = [from_classes]
    if not isinstance(to_classes, (tuple, list)):
        to_classes = [to_classes]

    for key in itertools.product(from_classes, to_classes):
        if key in __adapters__:
            raise AdapterExists('%r to %r already exists.' % key)
        __adapters__[key] = func